import uuid
from loguru import logger
from langchain.llms.base import BaseLLM
from langchain.agents import initialize_agent, AgentType
from lemonai.get_integrations import get_apis_from_env
from lemonai.api_wrapper import APIWrapper
from lemonai.toolkit import Toolkit
from lemonai.filter_tools import filter_tools
from typing import Optional


def execute_workflow(llm: BaseLLM, prompt_string: str, server_domain: Optional[str] = None):

    logfile_path = "lemonai.log"
    logger.remove(handler_id=None)
    logger.add(logfile_path, format="{time} - {extra[session_id]} - {extra[operation_name]}")

    api_keys_dict, access_tokens_dict = get_apis_from_env()
    session_id = uuid.uuid4()

    _wrapper = APIWrapper(llm, task=prompt_string, api_domain=server_domain)
    toolkit = Toolkit.from_api_wrapper(_wrapper, api_keys_dict, access_tokens_dict, logger, str(session_id))
    tools = toolkit.get_tools()

    prompt = f"Your task is '{prompt_string}'. Focus on the ordering of the tasks given. Do not use a workflow unless it is mentioned. Your action input should be a complete and valid JSON object where the keys are the params and the values are the value for each input parameter. The description of the tool may provide the JSON structure of the action input using round brackets () instead of curly brackets. Follow this structure for your action input. Ensure all strings in the action input are valid and terminated correctly. Your final answer should give a brief conversational overview of what you did."    
    filtered_tools = filter_tools(llm=llm, task=prompt_string, tools=tools)
    
    agent = initialize_agent(
        tools=filtered_tools,
        llm=llm,
        agent=AgentType.ZERO_SHOT_REACT_DESCRIPTION,
        verbose=True
    )

    agent.run(prompt)



import os
import json
from langchain.agents.agent_toolkits.base import BaseToolkit
from lemonai.api_wrapper import APIWrapper
from lemonai.tool import Tool
from lemonai.workflow import Workflow
from typing import List, Dict, Optional, Union, Tuple
from loguru._logger import Logger

class Toolkit(BaseToolkit):

    tools: List[Union[Tool, Workflow]] = []

    def _get_access_variable(self, api_key_dict: Dict[str, str], access_token_dict: Dict[str, str], action_id: str) -> Tuple[Optional[str], Optional[str]]:
        api_key = None
        access_token = None

        for key in api_key_dict:
            if key in action_id:
                api_key = api_key_dict[key]
        
        for key in access_token_dict:
            if key in action_id:
                access_token = access_token_dict[key]
    
        return api_key, access_token
    
    def _get_workflows_from_file(tools_list: List[Tool]) -> List[Workflow]:
        workflows = []

        with open("lemonai.json", "r") as file:
            data = json.load(file)
            tool_ids_list = [tool.id for tool in tools_list]

            for item in data:
                required_tools = item["tools"]
                tools_objs = []
                ignore = False
                for tool_id in required_tools:
                    index = tool_ids_list.index(tool_id) if tool_id in tool_ids_list else -1
                    if index == -1:
                        ignore = True
                        break
                    else:
                        tools_objs.append(tools_list[index])
                
                if not ignore:
                    tool_names = [tool.name for tool in tools_objs]
                    workflows.append(
                        Workflow(
                            name=item["name"],
                            description=Toolkit._build_workflow_description(item["name"], item["description"], tool_names),
                            tools=tools_objs,
                    ))

        return workflows

    def _build_workflow_description(name: str, description: str, tool_names: List[str]) -> str:
        return f"A wrapper around Lemon AI workflows. To run this workflow, run each tool in the workflow's tools list in the correct order to achieve the goal given by the workflow's description. You MUST take into account the params when executing each stage of the workflow. Sometimes, but not always, the output from a tool in the workflow may need to be used as the param for another tool in the workflow. For example, if the workflow's description is \"send the latest weather information in an email to maria.musterfrau@gmail.com\", first you would need to retrieve the latest weather information using one tool and then send that tool's output in an email to the given recipient, using the weather information as a param for the tool to send the email. You MUST use every tool that is given in the workflow's tools list once and in the same order as the list. Do not make up params, they will be explicitly specified in each tool's description. Do not use any tools that are not in this workflow's tools list. If you do not have enough information to fill in the params for a tool, just say 'not enough information provided in the instruction, missing <param>'. If you get a null or none response, STOP EXECUTION, do not try another tool! This workflow is named:{name} and has the goal:{description} and this workflow's tools list is: {tool_names}" 

    @classmethod
    def from_api_wrapper(
        cls,
        api_wrapper: APIWrapper,
        api_keys_dict: Dict[str, str],
        access_tokens_dict: Dict[str, str],
        logger: Logger,
        session_id: str,
    ) -> "Toolkit":
        
        actions = api_wrapper.get_tool_list()

        tools = []

        for action in actions:
            if action["authorizationType"] == "none":
                tools.append(Tool(
                    id=action["id"],
                    name=action["name"],
                    description=action["description"],
                    api_wrapper=api_wrapper,
                    logger=logger,
                    session_id=session_id
                ))
            else:
                api_key, access_token = cls._get_access_variable(cls, api_keys_dict, access_tokens_dict, action["id"])
                if api_key or access_token:
                    tools.append(Tool(
                        id=action["id"],
                        name=action["name"],
                        description=action["description"],
                        api_wrapper=api_wrapper,
                        api_key=api_key,
                        access_token=access_token,
                        logger=logger,
                        session_id=session_id
                    ))
        
        if os.path.exists("lemonai.json"):
            workflows = cls._get_workflows_from_file(tools)
            tools.extend(workflows)

        return cls(tools=tools)

    def get_tools(self) -> List[Tool]:
        return self.tools


import json
import requests
from requests import Request, Session
from typing import List, Dict, Optional
from langchain.llms.base import BaseLLM

class APIWrapper:

    api_base_url: str

    def __init__(self, llm: BaseLLM, task: str, api_domain: Optional[str] = None):
        self.api_base_url = f"{api_domain if api_domain else 'https://bfyvsp3cwepesc7dxd2yuzkyyi0nmcye.lambda-url.eu-central-1.on.aws'}/api/v0"
        self._llm = llm
        self._task = task

    api_base_url: str = "http://localhost:1313/api/v0/"
    _llm: BaseLLM
    _task: str

    def _get_session(self) -> Session:
        
        session = requests.session()
        session.headers.update(
            {
                "Accept": "application/json",
                "Content-Type": "application/json",
            }
        )

        return session

    def get_tool_list(self) -> List[Dict]:

        session = self._get_session()
        url = f"{self.api_base_url}/tools"
        response = session.get(url)
        response.raise_for_status()

        return response.json()
    
    def _get_action_request(
        self, id: str, action_input: str, params: Optional[Dict]
    ) -> Request:
        
        data = params if params else {}
        if action_input[-1] == ":":
            action_input = action_input + "\"\""
        json_obj = json.loads(action_input.replace("None", "\"\""))

        data.update(json_obj)

        return Request(
            method="POST",
            url=f"{self.api_base_url}/tool/{id}/execute",
            json=data,
        )

    def run(self, id: str, action_input: str, params: Optional[Dict], api_key: Optional[str], access_token: Optional[str]):
        
        session = self._get_session()
        request = self._get_action_request(id, action_input, params)

        if api_key:
            request.json["authToken"] = api_key
        elif access_token:
            request.json["authToken"] = access_token

        response = session.send(session.prepare_request(request))
        response.raise_for_status()

        json_list = response.json()

        if json_list:
            while isinstance(json_list, list):
                json_list = json_list[0]
            
            keys = list(json_list.keys())

            prompt = f"Based on the following task: '{self._task}', tell me which of the following keys: {keys} would be necessary to complete the task. Be as MINIMAL as possible. Focus on the entire task and specifically what values it is asking for. If the task mentions a specific key, this key MUST be in your final list. If the task does not specify what it is asking for, return the entire list of keys. Do not include any keys that are not in the given list. Your answer should be a valid LIST of keys (the minimum should be a list of 1 key) from the given list that would be necessary to complete the task."
            answer = self._llm.generate(prompts=[prompt]).generations[0][0].text
            list_string = answer.strip()

            key_list = eval(list_string)

            if isinstance(response.json(), list):
                json_list = response.json()

                flattened_list = []

                if isinstance(json_list[0], list):
                    for elem in json_list:
                        flattened_list.extend(elem)
                else:
                    flattened_list = json_list

                reduced_response = []
                for i in range(0, len(flattened_list)):
                    reduced_obj = {}
                    for key in key_list:
                        reduced_obj[key] = self._get_value(flattened_list[i], key)
                    
                    reduced_response.append(reduced_obj)
            else:
                reduced_response = {}
                for key in key_list:
                    reduced_response[key] = self._get_value(json_list, key)

            return reduced_response

    def _get_value(self, curr_dict, target_key):
        if isinstance(curr_dict, dict):
            for key, value in curr_dict.items():
                if key == target_key:
                    return value
                elif isinstance(value, dict):
                    result = self._get_value(value, target_key)
                    if result is not None:
                        return result
        return None



from langchain.tools import BaseTool
from lemonai.tool import Tool
from typing import List

class Workflow(BaseTool):

    name: str = ""
    description: str = ""
    tools: List[Tool] = []

    def _run(self, action_input: str) -> str:
        return "Need to run every tool in this workflow's tools list"

    async def _arun(self):
        raise NotImplementedError("Tool does not support async")


from langchain.tools import BaseTool
from pydantic import Field
from lemonai.api_wrapper import APIWrapper
from typing import Dict, Optional
from loguru._logger import Logger

class Tool(BaseTool):

    api_wrapper: APIWrapper = Field(default_factory=APIWrapper)
    id: str = ""
    name: str = ""
    description: str = ""
    params: Optional[Dict] = None
    api_key: Optional[str] = None
    access_token: Optional[str] = None
    logger: Logger
    session_id: str

    def _run(self, action_input: str) -> Dict:
        
        self.logger.bind(session_id=self.session_id, operation_name=self.id).info("")
        return self.api_wrapper.run(self.id, action_input, self.params, self.api_key, self.access_token)

    async def _arun(self):
        raise NotImplementedError(" Tool does not support async")


from langchain.llms.base import BaseLLM 
from lemonai.tool import Tool
from typing import List

def filter_tools(llm: BaseLLM, task: str, tools: List[Tool]) -> List[Tool]:

    tool_names = [tool.name for tool in tools]
    prompt = f"Given the following list of tools: {tool_names}, return a list of possible tools to be used to complete the following task: '{task}'. Consider only the tools given in the list. If the task mentions any workflow, you MUST include this workflow in your final answer. The only time you should suggest a workflow in your answer is if it explicitly stated in the task. You should split the task into subtasks and for each subtask, think about the possible tools that could be used to complete each subtask. Pay attention to tool names mentioned and verbs that are used (and their synonyms). Your answer should be a single valid list of the names of the tools that can be used to complete the given task."

    answer = llm.generate(prompts=[prompt]).generations[0][0].text
    list_string = answer.strip()

    filtered_tool_names = eval(list_string)

    filtered_tools = list(filter(lambda x: x.name in filtered_tool_names, tools))

    return filtered_tools


import os
from typing import Dict, Tuple

def get_apis_from_env() -> Tuple[Dict[str, str], Dict[str, str]]:

    api_key_search_strings = ['_API_KEY', '_SECRET_KEY', '_SUBSCRIPTION_KEY', '_ACCESS_KEY']
    access_token_search_strings = ['_ACCESS_TOKEN', '_SECRET_TOKEN', '_WEBHOOK_URL', '_API_TOKEN']

    api_keys = {}
    access_tokens = {}

    for key, value in os.environ.items():
        for search_string in api_key_search_strings:
            if search_string in key:
                name = key.split(search_string)[0].lower().replace("_", "-")
                api_keys[name] = value
        
        for search_string in access_token_search_strings:
            if search_string in key:
                name = key.split(search_string)[0].lower().replace("_", "-")
                access_tokens[name] = value

    return api_keys, access_tokens